SQ = "'"

COLUMN2LABEL = {
    f'Patient{SQ}s Birth Date': 'birth_date',
    f'Patient{SQ}s Sex': 'sex',
    f'Patient{SQ}s Age': 'age',
    f'Manufacturer{SQ}s Model Name': 'manufacturer_model',
    'Manufacturer': 'manufacturer',
    'Acquisition Date': 'acquisition_date',
    'Site': 'site',
}
